r"""Utility functions

"""

import numpy as onp
import jax.numpy as np
from typing import List


def C(n: int) -> np.ndarray:
    r"""The combinatorial :math:`C` matrix defined in the paper's appendix

    Args:
        n: the number of sampled haplotypes :math:`n`

    Returns:
        :math:`(n-1)\times(n-1)` matrix

    """
    W1 = np.zeros((n - 1, n - 1))
    W2 = np.zeros((n - 1, n - 1))
    b = np.arange(1, n - 1 + 1)
    # j = 2
    W1 = W1.at[:, 0].set(6 / (n + 1))
    W2 = W2.at[:, 0].set(0)
    # j = 3
    W1 = W1.at[:, 1].set(10 * (5 * n - 6 * b - 4) / (n + 2) / (n + 1))
    W2 = W2.at[:, 1].set((20 * (n - 2)) / (n+1) / (n+2))
    for col in range(n - 3):
        # this cast is crucial for floating point precision
        j = np.float64(col + 2)
        # procedurally generated by Zeilberger's algorithm in Mathematica
        W1 = W1.at[:, col + 2].set(-((-((-1 + j)*(1 + j)**2*(3 + 2*j)*(j - n)*(4 + 2*j - 2*b*j + j**2 - b*j**2 + 4*n + 2*j*n + j**2*n)*W1[:, col]) - (-1 + 2*j)*(3 + 2*j)*(-4*j - 12*b*j - 4*b**2*j - 6*j**2 - 12*b*j**2 - 2*b**2*j**2 - 4*j**3 + 4*b**2*j**3 - 2*j**4 + 2*b**2*j**4 + 4*n + 2*j*n - 6*b*j*n + j**2*n - 9*b*j**2*n - 2*j**3*n - 6*b*j**3*n - j**4*n - 3*b*j**4*n + 4*n**2 + 6*j*n**2 + 7*j**2*n**2 + 2*j**3*n**2 + j**4*n**2)*W1[:, col + 1])/(j**2*(2 + j)*(-1 + 2*j)*(1 + j + n)*(3 + b + j**2 - b*j**2 + 3*n + j**2*n))))  # noqa: E501
        W2 = W2.at[:, col + 2].set(((-1 + j)*(1 + j)*(2 + j)*(3 + 2*j)*(j - n)*(1 + j - n)*(1 + j + n)*W2[:, col] + (-1 + 2*j)*(3 + 2*j)*(1 + j - n)*(j + n)*(2 - j - 2*b*j - j**2 - 2*b*j**2 + 2*n + j*n + j**2*n)*W2[:, col + 1])/((-1 + j)*j*(2 + j)*(-1 + 2*j)*(j - n)*(j + n)*(1 + j + n)))   # noqa: E501

    return W1 - W2


def M(n: int, t: np.ndarray, y: np.ndarray) -> np.ndarray:
    r"""The M matrix defined in the paper's appendix

    Args:
        n: the number of sampled haplotypes :math:`n`
        t: time grid, starting at zero and ending at np.inf
        y: population size in each epoch

    Returns:
        :math:`(n-1)\times m` matrix, where :math:`m` is the number of epochs
        (the length of the ``y`` argument)

    """
    # epoch durations
    s = np.diff(t)
    # we handle the final infinite epoch carefully to facilitate autograd
    u = np.exp(-s[:-1] / y[:-1])
    u = np.concatenate((np.array([1]), u, np.array([0])))

    n_range = np.arange(2, n + 1)
    binom_vec = n_range * (n_range - 1) / 2

    return np.exp(binom_vec[:, np.newaxis]
                  * np.cumsum(np.log(u[np.newaxis, :-1]), axis=1)
                  - np.log(binom_vec[:, np.newaxis])) \
        @ (np.eye(len(y), k=0) - np.eye(len(y), k=-1)) \
        @ np.diag(y)


def tmrca_sf(t: np.ndarray, y: np.ndarray, n: int) -> np.ndarray:
    """The survival function of the TMRCA at each time point

    Args:
        t: time grid (including zero and infinity)
        y: effective population size in each epoch
        n: number of sampled haplotypes

    """
    # epoch durations
    s = np.diff(t)
    logu = -s / y
    logu = np.concatenate((np.array([0]), logu))
    # the A_2j are the product of this matrix
    # NOTE: using letter  "l" as a variable name to match text
    l = onp.arange(2, n + 1)[:, onp.newaxis]  # noqa: E741
    with onp.errstate(divide='ignore'):
        A2_terms = l * (l - 1) / (l * (l - 1) - l.T * (l.T - 1))
    onp.fill_diagonal(A2_terms, 1)
    A2 = np.prod(A2_terms, axis=0)

    binom_vec = l * (l - 1) / 2

    result = np.zeros(len(t))
    result = result.at[:-1].set(np.squeeze(A2[np.newaxis, :]
                                @ np.exp(np.cumsum(logu[np.newaxis, :-1],
                                                   axis=1)) ** binom_vec))

    assert np.all(np.isfinite(result))

    return result


complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}


def revcomp(seq: str) -> str:
    """reverse complement mutation type

    Args:
        seq: nucleotide string (all caps ACGT)
    """
    return ''.join(complement[b] for b in reversed(seq))


def misid_partners(mutation_types: List[str]) -> List[int]:
    """Ancestral state misidentification partner indices. Mutation type ``None``
    is self-partnered.

    Args:
        mutation_types: list of mutation type strings, e.g. [AAA>ACA, ...]

    """
    to_pair = list(enumerate(mutation_types))
    pair_idxs = [-1] * len(to_pair)
    while to_pair:
        i, mutype1 = to_pair[0]
        if mutype1 is None:
            pair_idxs[i] = i
            del to_pair[0]
        else:
            anc1, der1 = mutype1.split('>')
            match = False
            for new_idx, (j, mutype2) in enumerate(to_pair):
                anc2, der2 = mutype2.split('>')
                if (anc1, der1) in ((der2, anc2), (revcomp(der2), revcomp(anc2))):
                    match = True
                    break
            if not match:
                raise ValueError('no ancestral misidentification partner found '
                                 f'for mutation type {mutype1}')
            pair_idxs[i] = j
            pair_idxs[j] = i
            del to_pair[new_idx]
            if new_idx != 0:
                del to_pair[0]
    assert set(pair_idxs) == set(range(len(mutation_types)))
    return pair_idxs


def mutype_misid(mutation_types: List[str]):
    """mutation type misidentification operator

    Args:
        mutation_types: list of mutation type strings, e.g. [AAA>ACA, ...]
    """
    AM_mut = np.zeros((len(mutation_types), len(mutation_types)))
    for j, i in enumerate(misid_partners(mutation_types)):
        AM_mut = AM_mut.at[i, j].set(1)
    return AM_mut


def fold(x: np.ndarray) -> np.ndarray:
    """transform SFS to folded SFS"""
    """Loss under current history

    Args:
        func: loss function name from loss_functions module

    Returns:
        loss
    """
    n = len(x) + 1
    x = (x + x[::-1])[:(n // 2)]
    if n % 2 == 0:
        x = x.at[-1].set(x[-1] // 2)
    return x



r"""Optimizer objects

"""

import abc
import numpy as np
from typing import Callable, Tuple
import prox_tv as ptv
from scipy.linalg import cholesky_banded, cho_solve_banded
from functools import lru_cache


class Optimizer(metaclass=abc.ABCMeta):
    """Abstract base class for optimizers

    Attributes:
        x: solution point

    Args:
        verbose: flag to print convergence messages
    """

    def __init__(self, verbose: bool = False):
        self.verbose = verbose
        self.x = None

    @abc.abstractmethod
    def initialize(self, x: np.ndarray) -> None:
        """initialize solution point x, and any auxiliary variables"""
        pass

    def check_x(self):
        """test if x is defined"""
        if self.x is None:
            raise TypeError('solution point x is not initialized')

    @abc.abstractmethod
    def f(self) -> np.float64:
        """evaluate cost function at current solution point"""
        pass

    @abc.abstractmethod
    def step(self) -> None:
        """take an optimization step and update solution point"""
        pass

    def run(self, x: np.ndarray,
            tol: np.float64 = 1e-6, max_iter: int = 100) -> np.ndarray:
        """Optimize until convergence criteria are met

        Args:
            x: initial point
            tol: relative tolerance in objective function
            max_iter: maximum number of iterations

        Returns:
            x: solution point
        """
        self.initialize(x)
        self.check_x()
        # initial objective value
        f = self.f()
        if self.verbose:
            print(f'initial objective {f:.6e}', flush=True)
        k = 0
        for k in range(1, max_iter + 1):
            self.step()
            if not np.all(np.isfinite(self.x)):
                print('warning: x contains invalid values', flush=True)
            # terminate if objective function is constant within tolerance
            f_old = f
            f = self.f()
            rel_change = np.abs((f - f_old) / f_old)
            if self.verbose:
                print(f'iteration {k}, objective {f:.3e}, '
                      f'relative change {rel_change:.3e}',
                      end='        \r', flush=True)
            if rel_change < tol:
                if self.verbose:
                    print('\nrelative change in objective function '
                          f'{rel_change:.2g} '
                          f'is within tolerance {tol} after {k} iterations',
                          flush=True)
                return self.x
        if self.verbose and k > 0:
            print(f'\nmaximum iteration {max_iter} reached with relative '
                  f'change in objective function {rel_change:.2g}',
                  flush=True)
        return self.x


class LineSearcher(Optimizer):
    """Abstract class for an optimizer with Armijo line search

    Args:
        s0: initial step size
        max_line_iter: maximum number of line search steps
        gamma: step size shrinkage rate for line search
        verbose: flag to print convergence messages
    """

    def __init__(self, s0: np.float64 = 1, max_line_iter: int = 100,
                 gamma: np.float64 = 0.8, verbose: bool = False):
        self.s0 = s0
        self.max_line_iter = max_line_iter
        self.gamma = gamma
        super().__init__(verbose=verbose)


class AccProxGrad(LineSearcher):
    r"""Nesterov accelerated proximal gradient method with backtracking line
    search [1]_.

    The optimization problem solved is:

    .. math::
        \arg\min_x g(x) + h(x)

    where :math:`g` is differentiable, and the proximal operator for :math:`h`
    is available.

    Args:
        g: differentiable term in objective function
        grad_g: gradient of g
        h: non-differentiable term in objective function
        prox: proximal operator corresponding to h
        verbose: flag to print convergence messages
        line_search_kwargs: line search keyword arguments,
                            see :py:class:`mushi.optimization.LineSearcher`

    References
    ----------
    .. [1]
    https://people.eecs.berkeley.edu/~elghaoui/Teaching/EE227A/lecture18.pdf

    """

    def __init__(self,
                 g: Callable[[np.ndarray], np.float64],
                 grad_g: Callable[[np.ndarray], np.float64],
                 h: Callable[[np.ndarray], np.float64],
                 prox: Callable[[np.ndarray, np.float64], np.float64],
                 verbose: bool = False,
                 **line_search_kwargs):
        self.g = g
        self.grad_g = grad_g
        self.h = h
        self.prox = prox
        super().__init__(verbose=verbose, **line_search_kwargs)

    def f(self):
        self.check_x()
        return self.g(self.x) + self.h(self.x)

    def initialize(self, x: np.ndarray) -> None:
        # initialize solution point
        self.x = x
        # initialize momentum iterate
        self.q = self.x
        # initialize step size
        self.s = self.s0
        # initialize step counter
        self.k = 0

    def step(self) -> None:
        """step with backtracking line search"""
        self.check_x()
        # evaluate differtiable part of objective at momentum point
        g1 = self.g(self.q)
        grad_g1 = self.grad_g(self.q)
        if not np.all(np.isfinite(grad_g1)):
            raise RuntimeError(f'invalid gradient:\n{grad_g1}')
        # store old iterate
        x_old = self.x
        # Armijo line search
        for line_iter in range(self.max_line_iter):
            # new point via prox-gradient of momentum point
            self.x = self.prox(self.q - self.s * grad_g1, self.s)
            # G_s(q) as in the notes linked above
            G = (1 / self.s) * (self.q - self.x)
            # test g(q - sG_s(q)) for sufficient decrease
            if self.g(self.q - self.s * G) <= (
                                            g1 - self.s * (grad_g1 * G).sum()
                                            + (self.s / 2) * (G ** 2).sum()):
                # Armijo satisfied
                break
            else:
                # Armijo not satisfied
                self.s *= self.gamma  # shrink step size

        # update step count
        self.k += 1
        # update momentum point
        self.q = self.x + ((self.k - 1) / (self.k + 2)) * (self.x - x_old)

        if line_iter == self.max_line_iter - 1:
            print('warning: line search failed', flush=True)
            # reset step size
            self.s = self.s0


class ThreeOpProxGrad(AccProxGrad):
    r"""Three operator splitting proximal gradient method with backtracking
    line search [2]_.

    The optimization problem solved is:

    .. math::
        \arg\min_x g(x) + h_1(x) + h_2(x)

    where :math:`g` is differentiable, and the proximal operators for
    :math:`h_1` and :math:`h_2` are available.

    Args:
        g: differentiable term in objective function
        grad_g: gradient of g
        h1: 1st non-differentiable term in objective function
        h2: 2nd non-differentiable term in objective function
        prox1: proximal operator corresponding to h1
        prox2: proximal operator corresponding to h2
        verbose: print convergence messages
        line_search_kwargs: line search keyword arguments,
                            see :py:class:`mushi.optimization.LineSearcher`

    References:
        .. [2] Pedregosa, Gidel, Adaptive Three Operator Splitting in
               Proceedings of the 35th International Conference on Machine
               Learning, Proceedings of Machine Learning Research., J. Dy, A.
               Krause, Eds. (PMLR, 2018), pp. 4085–4094.

    """

    def __init__(self,
                 g: Callable[[np.ndarray], np.float64],
                 grad_g: Callable[[np.ndarray], np.float64],
                 h1: Callable[[np.ndarray], np.float64],
                 prox1: Callable[[np.ndarray, np.float64], np.float64],
                 h2: Callable[[np.ndarray], np.float64],
                 prox2: Callable[[np.ndarray, np.float64], np.float64],
                 verbose: bool = False,
                 **line_search_kwargs):
        super().__init__(g, grad_g, h1, prox1, verbose=verbose,
                         **line_search_kwargs)
        self.h2 = h2
        self.prox2 = prox2

    def f(self):
        return super().f() + self.h2(self.x)

    def initialize(self, x: np.ndarray) -> None:
        super().initialize(x)
        # dual variable
        self.u = np.zeros_like(self.q)

    def step(self) -> None:
        """step with backtracking line search"""
        self.check_x()
        # evaluate differentiable part of objective
        g1 = self.g(self.q)
        grad_g1 = self.grad_g(self.q)
        if not np.all(np.isfinite(grad_g1)):
            raise RuntimeError(f'invalid gradient:\n{grad_g1}')
        # Armijo line search
        for line_iter in range(self.max_line_iter):
            # new point via prox-gradient of momentum point
            self.x = self.prox(self.q - self.s * (self.u + grad_g1), self.s)
            # quadratic approximation of objective
            Q = (g1 + (grad_g1 * (self.x - self.q)).sum()
                    + ((self.x - self.q) ** 2).sum() / (2 * self.s))
            if self.g(self.x) - Q <= 0:
                # sufficient decrease satisfied
                break
            else:
                # sufficient decrease not satisfied
                self.s *= self.gamma  # shrink step size
        if line_iter == self.max_line_iter - 1:
            print('warning: line search failed', flush=True)
            # reset step size
            self.s = self.s0

        # update z variables with 2nd prox
        self.q = self.prox2(self.x + self.s * self.u, self.s)
        # update u variables: dual variables
        self.u = self.u + (self.x - self.q) / self.s
        # grow step size
        self.s = min(self.s / self.gamma ** 2, self.s0)


class TrendFilter(Optimizer):
    r"""Mixed trend filtering via specialized ADMM as in [3]_, section 5.2.

    The optimization problem solved is:

    .. math::
        \arg\min_{\beta \in \mathbb{R}} \frac{1}{2} \|y - \beta\|_2^2
        + \sum_{\ell=1}^r \lambda_\ell \| D^{k_\ell + 1} \beta \|_1

    where :math:`r` is the number of elements of k.

    Args:
        ks: tuple of integer trend filter orders
        lambdas: tuple of penalties corresponding to each k
        rhos: ADMM convergence parameter
        verbose: print convergence messages

    References:
        .. [3] Aaditya Ramdas and Ryan J. Tibshirani.
               Fast and flexible admm algorithms for trend filtering.
               Journal of Computational and Graphical Statistics
               25.3 (2016): 839-858.
    """

    def __init__(self, ks: Tuple[int], lambdas: Tuple[np.float64],
                 rhos: Tuple[np.float64] = None, verbose: bool = False):
        self.k = ks
        self.λ = lambdas
        self.r = len(self.k)
        if rhos is None:
            # default ADMM parameters used in cited paper
            self.ρ = self.λ
        else:
            self.ρ = rhos
        if len(self.λ) != self.r:
            raise ValueError('ks and lambdas must have equal length')
        if len(self.ρ) != self.r:
            raise ValueError('ks and rhos must have equal length')
        super().__init__(verbose=verbose)

    def f(self):
        self.check_x()
        fit = 0.5 * np.sum((self.x - self.y) ** 2)
        penalty = sum(λ * np.linalg.norm(np.diff(self.x, k + 1, axis=0), 1)
                      for k, λ in zip(self.k, self.λ))
        return fit + penalty

    def initialize(self, x: np.ndarray) -> None:
        # input signal
        self.y = x
        if self.y.ndim == 1:
            self.y = self.y[:, np.newaxis]
            # remember the signal was 1D
            self.oneD = True
        else:
            self.oneD = False
        self.n = self.y.shape[0]
        # initialize solution point (β in the paper cited)
        self.x = np.zeros_like(self.y)
        # initialize Lagrangian dual variables
        self.α = [np.zeros_like(self.y)[:(self.n - k)] for k in self.k]
        self.u = [np.zeros_like(self.y)[:(self.n - k)] for k in self.k]

        self.D, self.DTD = self._D_DTD(self.n, self.k)
        self.c = self._choleskify(self.n, self.k, self.ρ)

    def step(self):
        self.x = cho_solve_banded(
            (self.c, False),
            self.y + sum(self.ρ[i] * self.D[i].T @ (self.α[i] + self.u[i])
                         for i in range(self.r)),
            check_finite=False)
        for i in range(self.r):
            Dx = self.D[i] @ self.x
            for j in range(self.x.shape[1]):
                self.α[i][:, j] = ptv.tv1_1d(Dx[:, j] - self.u[i][:, j],
                                             self.λ[i] / self.ρ[i])
            self.u[i] += self.α[i] - Dx

    def run(self, *args, **kwargs) -> np.ndarray:
        solution = super().run(*args, **kwargs)
        # squeeze out singleton dimension if input was a 1D
        if self.oneD:
            return np.squeeze(solution)
        # else:
        return solution

    @staticmethod
    @lru_cache()
    def _D_DTD(n: int, k: Tuple[int]) -> Tuple[Tuple[np.ndarray],
                                               Tuple[np.ndarray]]:
        """difference operator for each order in k

        Args:
            n: number of points in signal
            k: tuple of integer trend filter orders

        Returns:
            D: tuple of difference operator for each k
            DTD: D.T @ D corresponding to above
        """
        D = np.eye(n, k=0) - np.eye(n, k=-1)
        D = tuple(np.linalg.matrix_power(D, k)[k:] for k in k)

        return tuple(zip(*((D[i], D[i].T @ D[i]) for i in range(len(k)))))

    @staticmethod
    @lru_cache(maxsize=1024)
    def _choleskify(n: int, k: Tuple[int], ρ: Tuple[np.float64]) -> np.ndarray:
        """cholesky decomposition needed for linear solves in trend estimate

        Args:
            n: number of points in signal
            k: tuple of integer trend filter orders
            ρ: ADMM parameter for each order in k

        Returns:
            Cholesky decomposition
        """
        DTD = TrendFilter._D_DTD(n, k)[1]
        A = np.eye(n) + sum(ρ[i] * DTD[i] for i in range(len(k)))
        # A is a banded Hermitian positive definite matrix with upper/lower
        # bandwidth bw. Express in upper diagonal ordered form
        bw = max(k)
        Ab = np.zeros((bw + 1, A.shape[1]))
        for u in range(bw + 1):
            Ab[-(1 + u), u:] = np.diag(A, k=u)

        return cholesky_banded(Ab, check_finite=False)



import mushi.histories as hst
from mushi import loss_functions, utils
import mushi.optimization as opt
import mushi.composition as cmp

from jax.config import config
import numpy as onp
import jax.numpy as np
from jax import jit, grad
from jax.scipy.special import expit, logit
from scipy.stats import poisson
from typing import Union, List, Dict, Tuple
from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
import pandas as pd
import seaborn as sns

config.update('jax_enable_x64', True)
# config.update('jax_debug_nans', True)


class kSFS():
    r"""Primary class for working with SFS data to infer demography
    :math:`\eta(t)\equiv 2N(t)`, or with :math:`k`-SFS data to infer demography
    and mutation spectrum history :math:`\boldsymbol\mu(t)`.

    Attributes:
        X (:obj:`numpy.ndarray`): :math:`k`-SFS matrix (or 1D SFS vector)
        eta (:obj:`mushi.eta`): demographic history
        mu (:obj:`mushi.mu`): mutation spectrum history
        mutation_types (:obj:`List[str]`): mutation spectrum history
        n (:obj:`int`): number of sampled haplotypes

    Notes:
        Three constructors:

        1) ``ksfs_file``: path to k-SFS file, as ouput by `mutyper ksfs`

        2) ``X`` and ``mutation_types`` (the latter may be ommitted if ``X`` is
           a 1D SFS array)

        3) ``n``: number of haplotypes to initialize for simulation

    Args:
        ksfs_file: path to :math:`k`-SFS file, as ouput by ``mutyper ksfs``
        X: :math:`k`-SFS matrix
        mutation_types: list of names of X columns
        n: number of sampled haplotypes

    """

    def __init__(self, X: np.ndarray = None, mutation_types: List[str] = None,
                 file: str = None,
                 n: int = None):
        if file is not None:
            df = pd.read_csv(file, sep='\t', index_col=0)
            assert np.all(df.values >= 0)
            n = df.shape[0] + 1
            self.X = np.array(df.values)  # NOTE: np is jax.numpy
            self.n = len(self.X) + 1
            self.mutation_types = pd.Index(df.columns,
                                           name='mutation type')

        elif X is not None:
            self.X = np.array(X)  # NOTE: np is jax.numpy
            self.n = len(X) + 1
            if self.X.ndim == 2:
                if mutation_types is None:
                    raise TypeError('must specify mutation_types')
                if len(mutation_types) != self.X.shape[1]:
                    raise ValueError('inconsistent number of mutation '
                                     f'types {len(mutation_types)} for X '
                                     f'with {self.X.shape[1]} columns')
                self.mutation_types = pd.Index(mutation_types,
                                               name='mutation type')
        elif n is None:
            raise TypeError('either file, or X, or n must be specified')
        else:
            self.n = n
            self.X = None
            self.mutation_types = None
        self.C = utils.C(self.n)
        self.η = None
        self.μ = None
        self.M = None
        self.L = None

        # ancestral state misidentification
        # misidentification rate
        self.r = None
        # mutation type-wise misidentification rate
        self.r_vector = None
        # frequency misidentification operator
        self.AM_freq = np.eye(self.n - 1)[::-1]
        # mutation type misidentification operator
        if self.X is not None and self.X.ndim == 2:
            self.AM_mut = utils.mutype_misid(self.mutation_types)

    @property
    def eta(self) -> hst.eta:
        r"""Read-only alias to η attribute
        """
        return self.η

    def check_eta(self):
        if self.η is None:
            raise TypeError('demographic history η is not defined')

    @property
    def mu(self) -> hst.mu:
        r"""Read-only alias to μ attribute
        """
        return self.μ

    def as_df(self) -> pd.DataFrame:
        r"""Return a pandas DataFrame representation
        """
        index = pd.Index(range(1, self.n), name='sample frequency')
        if self.X.ndim == 1:
            return pd.Series(self.X, index=index, name='SFS')
        elif self.X.ndim == 2:
            return pd.DataFrame(self.X, index=index,
                                columns=self.mutation_types)

    def clear_eta(self) -> None:
        r"""Clear demographic history attribute η
        """
        self.η = None
        self.M = None
        self.L = None

    def clear_mu(self) -> None:
        r"""Clear μ attribute
        """
        self.μ = None

    def tmrca_cdf(self, eta: hst.eta = None) -> onp.ndarray:
        r"""The CDF of the TMRCA at each change point in ``eta``

        Args:
            eta: demographic history (if ``None``, use ``self.eta`` attribute)
        """
        if eta is None:
            self.check_eta()
            eta = self.η
        t, y = eta.arrays()
        return 1 - utils.tmrca_sf(t, y, self.n)[1:-1]

    def simulate(self, eta: hst.eta, mu: Union[hst.mu, np.float64],
                 r: np.float64 = 0,
                 seed: int = None) -> None:
        r"""Simulate a :math:`k`-SFS under the Poisson random field model
        (no linkage), assign to ``X`` attribute

        Args:
            eta: demographic history
            mu: mutation spectrum history (or constant mutation rate)
            r: ancestral state misidentification rate (default 0)
            seed: random seed
        """
        onp.random.seed(seed)
        t, y = eta.arrays()
        M = utils.M(self.n, t, y)
        L = self.C @ M
        if type(mu) == hst.mu:
            eta.check_grid(mu)
            Ξ = L @ mu.Z
            self.mutation_types = mu.mutation_types
            self.AM_mut = utils.mutype_misid(self.mutation_types)
            self.X = np.array(poisson.rvs((1 - r) * Ξ
                                          + r * self.AM_freq @ Ξ @ self.AM_mut)
                              )
        else:
            ξ = mu * L.sum(1)
            self.X = np.array(poisson.rvs((1 - r) * ξ + r * self.AM_freq @ ξ))

    def set_eta(self, eta: hst.eta):
        r"""Set pre-specified demographic history :math:`\eta(t)`

        Args:
            eta: demographic history object
        """
        self.η = eta
        t = self.η.arrays()[0]
        self.M = utils.M(self.n, t, self.η.y)
        self.L = self.C @ self.M
        self.r = 0

    def infer_misid(self,
                    mu0: np.float64,
                    loss: str = 'prf',
                    max_iter: int = 100,
                    tol: np.float64 = 0,
                    line_search_kwargs: Dict = {},
                    verbose: bool = False
                    ) -> None:
        r"""Infer ancestral misidentification rate with :math:`\eta(t)` fixed.
        This function is used for fitting with a pre-specified demography, after
        using ``set_eta()``, instead of inferring :math:`\eta(t)` with
        ``infer_eta()`` (which jointly infers misidentification rate).

        Args:
            mu0: total mutation rate (per genome per generation)
            loss: loss function name from loss_functions module
            max_iter: maximum number of optimization steps
            tol: relative tolerance in objective function (if ``0``, not used)
            line_search_kwargs: line search keyword arguments,
                                see :py:meth:`mushi.optimization.LineSearcher`
            verbose: print verbose messages if ``True``
        """
        self.check_eta()
        if self.X is None:
            raise TypeError('use simulate() to generate data first')

        # total SFS
        if self.X.ndim == 1:
            x = self.X
        else:
            x = self.X.sum(1)

        # badness of fit
        loss = getattr(loss_functions, loss)

        self.mu0 = mu0

        @jit
        def g(r_logit):
            """differentiable piece of objective in η problem"""
            ξ = self.mu0 * self.L.sum(1)
            r = expit(r_logit)
            ξ = (1 - r) * ξ + r * self.AM_freq @ ξ
            return loss(np.squeeze(ξ), x)

        @jit
        def h(params):
            return 0

        @jit
        def prox(params, s):
            return params

        # optimizer
        optimizer = opt.AccProxGrad(g, jit(grad(g)), h, prox,
                                    verbose=verbose, **line_search_kwargs)
        # initial point
        r_logit = np.array([logit(1e-2)])
        # run optimization
        r_logit = optimizer.run(r_logit, tol=tol, max_iter=max_iter)

        self.r = expit(r_logit)

    def infer_eta(self,
                  mu0: np.float64,
                  *trend_penalty: Tuple[int, np.float64],
                  ridge_penalty: np.float64 = 0,
                  folded: bool = False,
                  pts: np.float64 = 100,
                  ta: np.float64 = None,
                  log_transform: bool = True,
                  eta: hst.eta = None,
                  eta_ref: hst.eta = None,
                  loss: str = 'prf',
                  max_iter: int = 100,
                  tol: np.float64 = 0,
                  line_search_kwargs: Dict = {},
                  trend_kwargs: Dict = {},
                  verbose: bool = False
                  ) -> None:
        r"""Infer demographic history :math:`\eta(t)`

        Args:
            mu0: total mutation rate (per genome per generation)
            trend_penalty: tuple ``(k, λ)`` for :math:`k`-th order trend
                           penalty (can pass multiple for mixed trends)
            ridge_penalty: ridge penalty
            folded: if ``False``, infer :math:`\eta(t)` using unfolded SFS. If
                    ``True``, can only be used with ``infer_mu=False``, and
                    infer :math:`\eta(t)` using folded SFS.
            pts: number of points for time discretization
            ta: time (in WF generations ago) of oldest change point in time
                discretization. If ``None``, set automatically based on
                10 * E[TMRCA] under MLE constant demography
            log_transform: fit :math:`\log\eta(t)`
            eta: initial demographic history. By default, a constant MLE is
                 computed
            eta_ref: reference demographic history for ridge penalty. If
                     ``None``, the constant MLE is used
            loss: loss function name from loss_functions module
            max_iter: maximum number of optimization steps
            tol: relative tolerance in objective function (if ``0``, not used)
            line_search_kwargs: line search keyword arguments,
                                see :py:meth:`mushi.optimization.LineSearcher`
            trend_kwargs: keyword arguments for trend filtering,
                          see :py:meth:`mushi.optimization.TrendFilter.run`
            verbose: print verbose messages if ``True``
        """
        if self.X is None:
            raise TypeError('use simulate() to generate data first')

        # total SFS
        if self.X.ndim == 1:
            x = self.X
        else:
            x = self.X.sum(1)
        # fold the spectrum if inference is on folded SFS
        if folded:
            x = utils.fold(x)

        # constant MLE
        # Harmonic number
        H = (1 / np.arange(1, self.n - 1)).sum()
        N_const = (self.X.sum() / 2 / H / mu0)

        if ta is None:
            tmrca_exp = 4 * N_const * (1 - 1 / self.n)
            ta = 10 * tmrca_exp
        change_points = np.logspace(0, np.log10(ta), pts)

        # ininitialize with MLE constant η
        if eta is not None:
            self.set_eta(eta)
        elif self.η is None:
            y = N_const * np.ones(change_points.size + 1)
            self.η = hst.eta(change_points, y)
        t = self.η.arrays()[0]

        self.mu0 = mu0

        # badness of fit
        loss = getattr(loss_functions, loss)

        # Accelerated proximal gradient method: our objective function
        # decomposes as f = g + h, where g is differentiable and h is not.
        # https://people.eecs.berkeley.edu/~elghaoui/Teaching/EE227A/lecture18.pdf

        # rescale trend penalties to be comparable between orders and time grids
        # filter zeros from trend penalties
        trend_penalty = tuple((k, (self.η.m ** k / onp.math.factorial(k)) * λ)
                              for k, λ in trend_penalty if λ > 0)

        # Tikhonov matrix
        if eta_ref is None:
            eta_ref = self.η
            Γ = np.diag(np.ones_like(eta_ref.y))
        else:
            self.η.check_grid(eta_ref)
            # - log(1 - CDF)
            Γ = np.diag(-np.log(utils.tmrca_sf(t, eta_ref.y, self.n))[:-1])
        y_ref = np.log(eta_ref.y) if log_transform else eta_ref.y

        # In the following, the parameter vector params will contain the
        # misid rate in params[0], and y in params[1:]
        @jit
        def g(params):
            """differentiable piece of objective in η problem"""
            y = params[1:]
            if log_transform:
                M = utils.M(self.n, t, np.exp(y))
            else:
                M = utils.M(self.n, t, y)
            L = self.C @ M
            ξ = self.mu0 * L.sum(1)
            if folded:
                ξ = utils.fold(ξ)
            else:
                r = expit(params[0])
                ξ = (1 - r) * ξ + r * self.AM_freq @ ξ
            loss_term = loss(np.squeeze(ξ), x)
            y_delta = y - y_ref
            ridge_term = (ridge_penalty / 2) * (y_delta.T @ Γ @ y_delta)
            return loss_term + ridge_term

        @jit
        def h(params):
            """nondifferentiable piece of objective in η problem"""
            return sum(λ * np.linalg.norm(np.diff(params[1:], k), 1)
                       for k, λ in trend_penalty)

        def prox(params, s):
            """trend filtering prox operator (no jit due to ptv module)"""
            if trend_penalty:
                k, sλ = zip(*((k, s * λ) for k, λ in trend_penalty))
                trend_filterer = opt.TrendFilter(k, sλ)
                params = params.at[1:].set(trend_filterer.run(params[1:],
                                                              **trend_kwargs))
            if log_transform:
                return params
            # else:
            # clip to minimum population size of 1
            return np.clip(params, 1)

        # optimizer
        optimizer = opt.AccProxGrad(g, jit(grad(g)), h, prox,
                                    verbose=verbose, **line_search_kwargs)
        # initial point
        params = np.concatenate((np.array([logit(1e-2)]),
                                 np.log(self.η.y)
                                 if log_transform else self.η.y))
        # run optimization
        params = optimizer.run(params, tol=tol, max_iter=max_iter)

        if not folded:
            self.r = expit(params[0])
        y = np.exp(params[1:]) if log_transform else params[1:]
        self.η = hst.eta(self.η.change_points, y)
        self.M = utils.M(self.n, t, y)
        self.L = self.C @ self.M

    def infer_mush(self,
                   *trend_penalty: Tuple[int, np.float64],
                   ridge_penalty: np.float64 = 0,
                   rank_penalty: np.float64 = 0,
                   hard: bool = False,
                   mu_ref: hst.mu = None,
                   misid_penalty: np.float64 = 1e-4,
                   loss: str = 'prf',
                   max_iter: int = 100,
                   tol: np.float64 = 0,
                   line_search_kwargs: Dict = {},
                   trend_kwargs: Dict = {},
                   verbose: bool = False
                   ) -> None:
        r"""Infer mutation spectrum history :math:`\mu(t)`

        Args:
            trend_penalty: tuple ``(k, λ)`` for :math:`k`-th order trend
                           penalty (can pass multiple for mixed trends)
            ridge_penalty: ridge penalty
            rank_penalty: rank penalty
            hard: hard rank penalty (non-convex)
            mu_ref: reference MuSH for ridge penalty. If None, the constant
                    MLE is used
            misid_penalty: ridge parameter to shrink misid rates to aggregate
                           rate
            loss: loss function from loss_functions module
            max_iter: maximum number of optimization steps
            tol: relative tolerance in objective function (if ``0``, not used)
            line_search_kwargs: line search keyword arguments,
                                see :py:class:`mushi.optimization.LineSearcher`
            trend_kwargs: keyword arguments for trend filtering,
                          see :py:meth:`mushi.optimization.TrendFilter.run`
            verbose: print verbose messages if ``True``
        """
        if self.X is None:
            raise TypeError('use simulate() to generate data first')
        self.check_eta()
        if self.r is None or self.r == 0:
            raise ValueError('ancestral misidentification rate has not been '
                             'inferred, possibly due to folded SFS inference')
        if self.mutation_types is None:
            raise ValueError('k-SFS must contain multiple mutation types')

        # number of segregating variants in each mutation type
        S = self.X.sum(0, keepdims=True)
        # ininitialize with MLE constant μ
        μ_const = hst.mu(self.η.change_points,
                         self.mu0 * (S / S.sum()) * np.ones((self.η.m,
                                                             self.X.shape[1])),
                         mutation_types=self.mutation_types.values)
        if self.μ is None:
            self.μ = μ_const
        t = μ_const.arrays()[0]
        self.M = utils.M(self.n, t, self.η.y)
        self.L = self.C @ self.M

        # badness of fit
        loss = getattr(loss_functions, loss)

        # rescale trend penalties to be comparable between orders and time grids
        # filter zeros from trend penalties
        trend_penalty = tuple((k, (self.μ.m ** k / onp.math.factorial(k)) * λ)
                              for k, λ in trend_penalty if λ > 0)

        if mu_ref is None:
            mu_ref = μ_const
            # Tikhonov matrix
            Γ = np.diag(np.ones_like(self.η.y))
        else:
            self.μ.check_grid(mu_ref)
            # - log(1 - CDF)
            Γ = np.diag(- np.log(utils.tmrca_sf(t, self.η.y, self.n))[:-1])

        # orthonormal basis for Aitchison simplex
        # NOTE: instead of Gram-Schmidt could try SVD of clr transformed X
        #       https://en.wikipedia.org/wiki/Compositional_data#Isometric_logratio_transform
        basis = cmp._gram_schmidt_basis(self.μ.Z.shape[1])
        check_orth = True if self.μ.Z.shape[1] > 2 else False

        # constand MLE and reference mush
        Z_const = cmp.ilr(μ_const.Z, basis, check_orth)
        Z_ref = cmp.ilr(mu_ref.Z, basis, check_orth)

        # weights for relating misid rates to aggregate misid rate from eta step
        misid_weights = self.X.sum(0) / self.X.sum()
        # reference composition for weighted misid (if all rates are equal)
        misid_ref = cmp.ilr(misid_weights, basis, check_orth)

        # In the following, params will hold the weighted misid composition in
        # the first row and the mush composition at each time in the remaining rows

        @jit
        def g(params):
            """differentiable piece of objective in μ problem"""
            r = self.r * cmp.ilr_inv(params[0, :], basis) / misid_weights
            Z = params[1:, :]
            Ξ = self.L @ (self.mu0 * cmp.ilr_inv(Z, basis))
            Ξ =  Ξ * (1 - r) + self.AM_freq @ Ξ @ (self.AM_mut * r[:, np.newaxis])
            loss_term = loss(Ξ, self.X)
            Z_delta = Z - Z_ref
            ridge_term = (ridge_penalty / 2) * np.sum(Z_delta * (Γ @ Z_delta))
            misid_delta = params[0, :] - misid_ref
            misid_ridge_term = misid_penalty * np.sum(misid_delta ** 2)
            return loss_term + ridge_term + misid_ridge_term

        if trend_penalty:

            @jit
            def h_trend(params):
                """trend filtering penalty"""
                return sum(λ * np.linalg.norm(np.diff(params[1:, :], k,
                                                      axis=0), 1)
                           for k, λ in trend_penalty)

            def prox_trend(params, s):
                """trend filtering prox operator (no jit due to ptv module)"""
                k, sλ = zip(*((k, s * λ) for k, λ in trend_penalty))
                trend_filterer = opt.TrendFilter(k, sλ)
                return params.at[1:, :].set(trend_filterer.run(params[1:, :],
                                              **trend_kwargs))

        if rank_penalty:
            if self.mutation_types.size < 3:
                raise ValueError('kSFS must have more than 2 mutation types for'
                                 ' rank penalization')

            @jit
            def h_rank(params):
                """2nd nondifferentiable piece of objective in μ problem"""
                if hard:
                    return rank_penalty * np.linalg.matrix_rank(params[1:, :]
                                                                - Z_const)
                # else:
                return rank_penalty * np.linalg.norm(params[1:, :] - Z_const,
                                                     'nuc')

            def prox_rank(params, s):
                """singular value thresholding"""
                U, σ, Vt = np.linalg.svd(params[1:, :] - Z_const,
                                         full_matrices=False)
                if hard:
                    σ = σ.at[σ <= s * rank_penalty].set(0)
                else:
                    σ = np.maximum(0, σ - s * rank_penalty)
                Σ = np.diag(σ)
                return params.at[1:, :].set(Z_const + U @ Σ @ Vt)

            if not hard:
                prox_rank = jit(prox_rank)

        # optimizer
        if trend_penalty and rank_penalty:
            optimizer = opt.ThreeOpProxGrad(g, jit(grad(g)),
                                            h_trend, prox_trend,
                                            h_rank, prox_rank,
                                            verbose=verbose,
                                            **line_search_kwargs)
        else:
            if trend_penalty:
                h = h_trend
                prox = prox_trend
            elif rank_penalty:
                h = h_rank
                prox = prox_rank
            else:
                @jit
                def h(params):
                    return 0

                @jit
                def prox(params, s):
                    return params

            optimizer = opt.AccProxGrad(g, jit(grad(g)), h, prox,
                                        verbose=verbose, **line_search_kwargs)

        # initial point (note initial row is for misid rates)
        # ---------------------------------------------------
        params = np.zeros((self.μ.m + 1, self.mutation_types.size - 1))
        # misid rate for each mutation type
        if self.r_vector is not None:
            r = self.r_vector
        else:
            r = self.r * np.ones(self.mutation_types.size)
        params = params.at[0, :].set(cmp.ilr(misid_weights * r, basis, check_orth))
        # ilr transformed mush
        ilr_mush = cmp.ilr(self.μ.Z, basis, check_orth)
        # make sure it's a column vector if only 2 mutation types
        if ilr_mush.ndim == 1:
            ilr_mush = ilr_mush[:, np.newaxis]
        params = params.at[1:, :].set(ilr_mush)
        # ---------------------------------------------------

        # run optimizer
        params = optimizer.run(params, tol=tol, max_iter=max_iter)

        # update attributes
        self.r_vector = self.r * cmp.ilr_inv(params[0, :], basis) / misid_weights
        self.μ = hst.mu(self.η.change_points,
                        self.mu0 * cmp.ilr_inv(params[1:, :], basis),
                        mutation_types=self.mutation_types.values)

    def plot_total(self, kwargs: Dict = dict(ls='', marker='.'),
                   line_kwargs: Dict = dict(),
                   fill_kwargs: Dict = dict(),
                   folded: bool = False) -> None:
        r"""Plot the SFS using matplotlib

        Args:
            kwargs: keyword arguments for scatter plot
            line_kwargs: keyword arguments for expectation line
            fill_kwargs: keyword arguments for marginal fill
            folded: if ``True``, plot the folded SFS and fit
        """
        if self.X.ndim == 1:
            x = self.X
        else:
            x = self.X.sum(1)
        if folded:
            x = utils.fold(x)
        plt.plot(range(1, len(x) + 1), x, **kwargs)
        if self.η is not None:
            if 'label' in kwargs:
                del kwargs['label']
            if self.μ is not None:
                self.η.check_grid(self.μ)
                z = self.μ.Z.sum(1)
            else:
                z = self.mu0 * np.ones_like(self.η.y)
            ξ = self.L.dot(z)
            if folded:
                ξ = utils.fold(ξ)
            else:
                if self.r is None:
                    raise TypeError('ancestral state misidentification rate '
                                    'is not inferred, do you want '
                                    'folded=True?')
                ξ = (1 - self.r) * ξ + self.r * self.AM_freq @ ξ
            plt.plot(range(1, len(ξ) + 1), ξ, **line_kwargs)
            ξ_lower = poisson.ppf(.025, ξ)
            ξ_upper = poisson.ppf(.975, ξ)
            plt.fill_between(range(1, len(ξ) + 1),
                             ξ_lower, ξ_upper, **fill_kwargs)
        plt.xlabel('sample frequency')
        plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
        plt.ylabel(r'variant count')
        plt.tight_layout()

    def plot(self, types=None, clr: bool = False,
             kwargs: Dict = dict(ls='', marker='.', rasterized=True),
             line_kwargs: Dict = dict()) -> None:
        r"""Plot the :math:`k`-SFS

        Args:
            types: iterable of mutation type names to restrict plotting to
            clr: flag to normalize to total mutation intensity and display as
                 centered log ratio transform
            kwargs: key word arguments passed to data scatter plot
            line_kwargs: key word arguments passed to expectation line plot
        """
        if self.η is not None and self.r is None:
            print('warning: misidentification rate is not defined, perhaps due'
                  ' to folded SFS inference, and will be set to 0 for plotting')
            r = 0
        else:
            r = self.r
        if self.μ is not None:
            Ξ = self.L @ self.μ.Z
            Ξ = Ξ * (1 - self.r_vector) + self.AM_freq @ Ξ @ (self.AM_mut * self.r_vector[:, np.newaxis])
        if clr:
            X = cmp.clr(self.X)
            if self.μ is not None:
                Ξ = cmp.clr(Ξ)
            plt.ylabel('variant count composition\n(CLR transformed)')
        else:
            X = self.X
            plt.ylabel('number of variants')
        if types is not None:
            idxs = [self.mutation_types.get_loc(type) for type in types]
            X = X[:, idxs]
            if self.μ is not None:
                Ξ = Ξ[:, idxs]

        plt.plot(range(1, self.n), X, **kwargs)
        if self.μ is not None:
            plt.gca().set_prop_cycle(None)
            plt.plot(range(1, self.n), Ξ, **line_kwargs)
        plt.xlabel('sample frequency')
        plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
        plt.xscale('log')
        plt.tight_layout()

    def clustermap(self, **kwargs) -> None:
        r"""Clustermap of compositionally centralized k-SFS

        Args:
            kwargs: additional keyword arguments passed to pandas.clustermap
        """
        Z = cmp.centralize(self.X)
        cbar_label = 'variant count\nperturbation'
        df = pd.DataFrame(data=Z, index=pd.Index(range(1, self.n),
                                                 name='sample frequency'),
                          columns=self.mutation_types)
        g = sns.clustermap(df, row_cluster=False,
                           center=1 / Z.shape[1],
                           cbar_kws={'label': cbar_label}, **kwargs)
        g.ax_heatmap.set_yscale('symlog')
        g.ax_heatmap.set_xticklabels(g.ax_heatmap.get_xmajorticklabels(),
                                     fontsize=9, family='monospace')

    def loss(self, func='prf') -> onp.float64:
        """Loss under current history

        Args:
            func: loss function name from loss_functions module

        Returns:
            loss
        """
        self.check_eta()
        loss = getattr(loss_functions, func)
        if self.μ is None:
            ξ = self.mu0 * self.L.sum(1)
            if self.r is not None:
                ξ = utils.fold(ξ)
                x = utils.fold(self.X.sum(1))
            else:
                ξ = (1 - self.r) * ξ + self.r * self.AM_freq @ ξ
            return loss(np.squeeze(ξ), np.squeeze(x))
        # else:
        Ξ = self.L @ self.μ.Z
        Ξ = Ξ * (1 - self.r_vector) + self.AM_freq @ Ξ @ (self.AM_mut * self.r_vector[:, np.newaxis])
        return onp.float64(loss(Ξ, self.X))



"""Loss functions"""

import jax.numpy as np


def prf(E: np.ndarray, X: np.ndarray) -> np.float64:
    r"""Poisson random field loss

    Args:
        E: expectation :math:`E[X]`
        X: data

    """
    return (E - X * np.log(E)).sum()


def dkl(E: np.ndarray, X: np.ndarray) -> np.float64:
    r"""generalized Kullback-Liebler divergence, a Bregman divergence (ignores
    constant term)

    Args:
        E: expectation :math:`E[X]`
        X: data

    """
    return (X * np.log(X / E) - X + E).sum()


def lsq(E: np.ndarray, X: np.ndarray) -> np.float64:
    r"""Least-squares loss

    Args:
        E: expectation :math:`E[X]`
        X: data

    """
    return (1 / 2) * ((E - X) ** 2).sum()



r"""History objects for representing demography and MuSH

"""

from dataclasses import dataclass
from typing import List

import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
import mushi.composition as cmp


@dataclass
class History():
    """
    Base class piecewise constant history. The first epoch starts at zero,
    and the last epoch extends to infinity

    Args:
        change_points: epoch change points (times)
        vals: constant values for each epoch (rows)
    """
    change_points: np.array
    vals: np.ndarray

    def __post_init__(self):
        if any(np.diff(self.change_points) <= 0) or any(
           np.isinf(self.change_points)) or any(self.change_points <= 0):
            raise ValueError('change_points must be increasing, finite, and '
                             'positive')
        if len(self.vals) != len(self.change_points) + 1:
            raise ValueError(f'len(change_points) = {len(self.change_points)}'
                             f' implies {len(self.change_points) + 1} epochs,'
                             f' but len(vals) = {len(self.vals)}')
        if np.any(self.vals < 0) or np.sum(np.isinf(self.vals)):
            raise ValueError('elements of vals must be finite and nonnegative')
        self.m = len(self.vals)

    def arrays(self):
        """Return time grid and values in each cell"""
        t = np.concatenate((np.array([0]),
                            self.change_points,
                            np.array([np.inf])))
        return t, self.vals

    def epochs(self):
        """Generator yielding epochs as tuples: (start_time, end_time, value)
        """
        for i in range(self.m):
            if i == 0:
                start_time = 0
            else:
                start_time = self.change_points[i - 1]
            if i == self.m - 1:
                end_time = np.inf
            else:
                end_time = self.change_points[i]
            value = self.vals[i]
            yield (start_time, end_time, value)

    def check_grid(self, other: 'History'):
        """Test if time grid is the same as another instance

        Args:
            other: another History object
        """
        if not np.array_equal(self.change_points, other.change_points):
            raise ValueError('histories have different time grids')

    def plot(self, t_gen: float = None, types=None,
             **kwargs) -> List[mpl.lines.Line2D]:
        """Plot the history

        Args:
            t_gen: generation time in years (optional)
            kwargs: key word arguments passed to plt.plot
        """
        t = np.concatenate((np.array([0]), self.change_points))
        if t_gen:
            t *= t_gen
            t_unit = 'years ago'
        else:
            t_unit = 'generations ago'
        if types is not None:
            idxs = [self.mutation_types.get_loc(type) for type in types]
            vals = self.vals[:, idxs]
        else:
            vals = self.vals
        lines = plt.plot(t, vals, **kwargs)
        plt.xlabel(f'$t$ ({t_unit})')
        plt.xscale('log')
        plt.tight_layout()
        return lines


class eta(History):
    """Demographic history

    Args:
        change_points: epoch change points (times)
        y: vector of constant population sizes in each epoch
    """
    @property
    def y(self):
        """read-only alias to vals attribute in base class"""
        return self.vals

    @y.setter
    def y(self, value):
        self.vals = value

    def __post_init__(self):
        super().__post_init__()
        assert len(self.y.shape) == 1, self.y.shape

    def plot(self, **kwargs) -> None:
        super().plot(**kwargs)
        plt.ylabel('$\\eta(t)$')
        plt.yscale('log')
        plt.tight_layout()


@dataclass
class mu(History):
    """Mutation spectrum history (MuSH)

    Args:
        change_points: epoch change points (times)
        Z: matrix of constant values for each epoch (rows) in each mutation type
           (columns)
           mutation_types: list of mutation type names (optional ``None`` element)
    """
    mutation_types: List[str]

    @property
    def Z(self):
        """read-only alias to vals attribute in base class"""
        return self.vals

    @Z.setter
    def Z(self, value):
        self.vals = value

    def __post_init__(self):
        super().__post_init__()
        # if mutation rate vector instead of matrix, promote to matrix
        if len(self.Z.shape) == 1:
            self.Z = self.Z[:, np.newaxis]
        assert len(self.Z.shape) == 2, self.Z.shape
        assert len(self.mutation_types) == self.Z.shape[1]
        self.mutation_types = pd.Index(self.mutation_types,
                                       name='mutation type')

    def plot(self, types: List[str] = None, clr: bool = False,
             **kwargs) -> None:
        """Plot history

        Args:
            types: list of mutation types to plot (default all)
            clr: flag to normalize to total mutation intensity and display as
                 centered log ratio transform
        """
        lines = super().plot(types=types, **kwargs)
        if clr:
            Z = cmp.clr(self.Z)
            if types is None:
                idxs = range(len(lines))
            else:
                idxs = [self.mutation_types.get_loc(type) for type in types]
            for idx, line in zip(idxs, lines):
                line.set_ydata(Z[:, idx])
            # recompute the ax.dataLim
            ax = plt.gca()
            ax.relim()
            # update ax.viewLim using the new dataLim
            ax.autoscale_view()
            # plt.ylabel('relative mutation intensity')
            plt.ylabel('mutation intensity composition\n(CLR transformed)')
        else:
            plt.ylabel('$\\mu(t)$')
        plt.tight_layout()

    def plot_cumulative(self, t_gen: float = None, clr: bool = False,
                        **kwargs) -> None:
        r"""Plot the cumulative mutation rate, like a Muller plot

        Args:
            t_gen: generation time in years (optional)
            clr: flag to normalize to total mutation intensity and display as
                 centered log ratio transform
            kwargs: key word arguments passed to ``plt.fill_between``
        """
        t = np.concatenate((np.array([0]), self.change_points))
        if t_gen:
            t *= t_gen
            t_unit = 'years ago'
        else:
            t_unit = 'generations ago'
        Z = np.cumsum(self.Z, axis=1)
        if clr:
            Z = cmp.clr(Z)
        for j in range(Z.shape[1]):
            plt.fill_between(t, Z[:, j - 1] if j else 0, Z[:, j], **kwargs)
        plt.xlabel(f'$t$ ({t_unit})')
        plt.ylabel('$\\mu(t)$')
        plt.xscale('log')
        plt.tight_layout()

    def clustermap(self, t_gen: float = None, **kwargs) -> None:
        r"""Clustermap of compositionally centralized MuSH

        Args:
            t_gen: generation time in years (optional)
            kwargs: additional keyword arguments passed to ``pd.clustermap``
        """
        t = np.concatenate((np.array([0]), self.change_points))
        if t_gen:
            t *= t_gen
            t_unit = 'years ago'
        else:
            t_unit = 'generations ago'
        Z = cmp.centralize(self.Z)
        label = 'mutation intensity\nperturbation'
        df = pd.DataFrame(data=Z, index=pd.Index(t, name=f'$t$ ({t_unit})'),
                          columns=self.mutation_types)
        g = sns.clustermap(df, row_cluster=False,
                           cbar_kws={'label': label},
                           **kwargs)
        g.ax_heatmap.set_xticklabels(g.ax_heatmap.get_xmajorticklabels(),
                                     fontsize=9, family='monospace')
